"""library of function to be used for working with time"""

from dateutil import parser
import datetime
from omigo_core import utils

def datetime_to_utctimestamp_millis(x):
    # convert this to string first for failsafe
    x = str(x)

    # 1681202675933
    if (len(str(x)) == 13 and str(x).isnumeric() == True):
        # this looks like numeric timestamp in millis
        return int(x)

    # 1681202675.933
    if (len(str(x)) == 14 and str(x).find(".") == 10 and str(x).isnumeric() == True):
        # this looks like numeric timestamp in millis
        return int(x)

    # 2023-04-11T08:44:35.933Z
    if (len(x) == 24 and x[19] == "." and x.endswith("Z")):
        return int(float(parser.parse(x).timestamp() * 1000))

    # 2023-04-15T15:05:16.175000Z
    if (len(x) == 27 and x[19] == "." and x.endswith("Z")):
        return int(float(parser.parse(x).timestamp() * 1000))

    # 2023-04-11T08:44:35.933+00:00
    if (len(x) == 29 and x[-6] == "+"):
        return int(float(parser.parse(x).timestamp() * 1000))

    # 2023-04-18T18:47:45 or 2023-04-18 18:47:45
    if (len(x) == 19 and (x[10] == "T" or x[10] == " ")):
        return int(float(parser.parse(x + "+00:00").timestamp() * 1000))

    # this seems to be a timestamp with second precision.
    return int(datetime_to_utctimestamp_sec(x) * 1000)

# TODO. better naming
def datetime_to_utctimestamp(x):
    # use datetime_to_utctimestamp_sec
    utils.warn_once("datetime_to_utctimestamp: Deprecated. Use datetime_to_utctimestamp_sec instead")
    return datetime_to_utctimestamp_sec(x)

def datetime_to_utctimestamp_sec(x):
    # convert this to string first for failsafe
    x = str(x)

    # 2022-05-20T05:00:00+00:00
    if (x.endswith("UTC") or x.endswith("GMT") or x.endswith("Z") or x.endswith("+00:00")):
        return int(parser.parse(x).timestamp())
    elif (len(x) == 10 and x.find("-") != -1):
        # 2021-11-01
        x = x + "T00:00:00Z"
        return int(parser.parse(x).timestamp())
    elif (len(x) == 19):
        # 2021-11-01T00:00:00
        x = x + "Z"
        if (x[10] == " "):
            x = x.replace(" ", "T")
        return int(parser.parse(x).timestamp())
    elif (len(x) == 23 and x[19] == "."):
        # 2021-11-01T00:00:00.000
        x = x + "Z"
        return int(parser.parse(x).timestamp())
    elif (len(x) == 26):
        # 2021-11-01T00:00:00.000000
        x = x + "Z"
        return int(parser.parse(x).timestamp())
    elif (len(x) == 27 and x[19] == "." and x.endswith("Z")):
        # 2023-04-15T15:05:16.175000Z
        return int(parser.parse(x).timestamp())
    elif (len(x) == 29 and x[-6] == "+"):
        # 2023-04-11T08:44:35.933+00:00
        return int(parser.parse(x).timestamp())
    elif (len(x) == 10 and str(x).isnumeric() == True):
        # this looks like a numeric timestamp
        return int(x)
    elif (len(x) == 13 and str(x).isnumeric() == True):
        # this looks like numeric timestamp in millis
        return int(int(x) / 1000)
    elif (len(x) == 19 and (x[10] == "T" or x[10] == " ")):
        # 2023-04-18T18:47:45 or 2023-04-18 18:47:45
        return int(float(parser.parse(x + "+00:00").timestamp() * 1000))
    else:
        raise Exception("Unknown date format. Problem with UTC: '{}'".format(x))

# TODO: Converts seconds format only. Even the original time in milliseconds will return seconds format
def utctimestamp_to_datetime(x):
    # use the string form
    x = str(x)
    if (len(x) == 10 and x.isnumeric() == True):
        return datetime.datetime.utcfromtimestamp(int(x)).replace(tzinfo = datetime.timezone.utc)
    elif (len(x) == 13 and x.isnumeric() == True):
        return datetime.datetime.utcfromtimestamp(int(x)/1000).replace(tzinfo = datetime.timezone.utc)
    elif (len(x) > 10 and x.find(".") == 10 and utils.is_float(x)): 
        return datetime.datetime.utcfromtimestamp(float(x)).replace(tzinfo = datetime.timezone.utc)
    else:
        raise Exception("Unknown timestamp format: {}".format(x))

# TODO: Converts seconds format only
def utctimestamp_millis_to_datetime(x):
    return utctimestamp_to_datetime(x)

# TODO: Converts seconds format only
# Its utc so removed the last timezone
def utctimestamp_to_datetime_str(x):
    return utctimestamp_to_datetime(x).isoformat()[0:19]

# Its utc so removed the last timezone. TODO: Keep the UTC or Z
def utctimestamp_millis_to_datetime_str(x):
    result = utctimestamp_to_datetime(x).isoformat()
    if (result.endswith("UTC")):
        return result[0:23]
    else:
        return result

def datetime_to_timestamp(x):
    raise Exception("Please use datetime_to_utctimestamp")

def get_utctimestamp_sec():
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp())

def get_utctimestamp_millis():
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1000)

def datestr_to_datetime(x):
    return utctimestamp_to_datetime(datetime_to_utctimestamp_sec(x))




from omigo_hydra import s3io_wrapper
from omigo_core import utils

class S3FSWrapper:
    def __init__(self):
        self.fs = s3io_wrapper.S3FSWrapper() 

    def file_exists(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.file_exists(*args, **kwargs)

    def file_exists_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.file_exists_with_wait(*args, **kwargs)

    def dir_exists(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.dir_exists(*args, **kwargs)

    def dir_exists_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.dir_exists_with_wait(*args, **kwargs)

    def file_not_exists(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.file_not_exists(*args, **kwargs)

    def file_not_exists_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.file_not_exists_with_wait(*args, **kwargs)

    def dir_not_exists_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.dir_not_exists_with_wait(*args, **kwargs)

    def read_file_contents_as_text_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.read_file_contents_as_text_with_wait(*args, **kwargs)

    def read_file_contents_as_text(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.read_file_contents_as_text(*args, **kwargs)

    def ls(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.ls(*args, **kwargs)

    def get_directory_listing(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.get_directory_listing(*args, **kwargs)

    def list_dirs(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.list_dirs(*args, **kwargs)

    def list_files(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.list_files(*args, **kwargs)

    def is_file(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.is_file(*args, **kwargs)

    def is_directory(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.is_directory(*args, **kwargs)

    def delete_file_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.delete_file_with_wait(*args, **kwargs)

    def delete_file(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.delete_file(*args, **kwargs)

    def delete_dir_with_wait(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.delete_dir_with_wait(*args, **kwargs)

    def get_parent_directory(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.get_parent_directory(*args, **kwargs)

    def write_text_file(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.write_text_file(*args, **kwargs)

    def create_dir(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.create_dir(*args, **kwargs)

    def makedirs(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.makedirs(*args, **kwargs)

    def get_last_modified_timestamp(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.get_last_modified_timestamp(*args, **kwargs)

    def copy_leaf_dir(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.copy_leaf_dir(*args, **kwargs)

    def list_leaf_dir(self, *args, **kwargs):
        utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra")
        return self.fs.list_leaf_dir(*args, **kwargs)





import unittest
from omigo_core import tsv

class TestTSV(unittest.TestCase):
    def create_tsv1(self):
        return tsv.TSV("name",["x"])

    def test_select1(self):
        xtsv = self.create_tsv1()
        self.assertEqual(xtsv.select("name").columns(), ["name"])

    def test_select2(self):
        xtsv = self.create_tsv1()
        self.assertEqual(xtsv.select("n.*").columns(), ["name"])

if __name__ == '__main__':
    unittest.main()



from omigo_core import tsv, utils, funclib
import queue

# This is WIP. This detects cycles and ignore assigning them levels
# TODO: there is this reverse_first flag which can be confusing. This api needs to be broken down into
# forward or reverse only
def get_bfs_levels(etsv, vertex_ids, reverse_first = True):
    # create map of levels
    fwd_levels = {}
    found_ids = []
    for v in vertex_ids:
        fwd_levels[v] = 0
        found_ids.append(v)

    # list of ids found so far in crawl
    # found_ids = [vertex_id]
    flag = True

    # run loop
    current_level = 0
    while (flag == True):
        new_flag = False
        for src, dest in etsv.to_tuples(["src", "target"]):
            # check if it is a new id
            if (src in found_ids and dest not in found_ids and fwd_levels[src] <= current_level):
                fwd_levels[dest] = fwd_levels[src] + 1
                found_ids.append(dest)
                new_flag = True

        # set flag if next loop is to be run
        if (new_flag == False):
            flag = False

        # increase level
        current_level = current_level + 1

    # run second round for reverse
    rev_levels = {}
    found_ids = []
    for v in vertex_ids:
        rev_levels[v] = 0
        found_ids.append(v)

    flag = True

    # run loop
    current_level = 0
    while (flag == True):
        new_flag = False
        for src, dest in etsv.to_tuples(["target", "src"]):
            # check if it is a new id
            if (src in found_ids and dest not in found_ids and rev_levels[src] >= current_level):
                # if level is not mapped already
                rev_levels[dest] = rev_levels[src] - 1
                found_ids.append(dest)
                new_flag = True

        # set flag if next loop is to be run
        if (new_flag == False):
            flag = False

        # decrease level
        current_level = current_level - 1

    # check in which order the levels are to be picked
    levels = {}
    combined_levels = [rev_levels, fwd_levels] if (reverse_first == True) else [fwd_levels, rev_levels]

    # assign if not assigned already
    for xlevels in combined_levels:
        for k in xlevels.keys():
            if (k not in levels.keys()):
                levels[str(k)] = xlevels[k]

    # return
    return levels

def get_forward_edges_only(etsv, prefix, sep = ","):
    # list of ids found so far in crawl
    paths = {}
    all_paths = {}

    # initialize
    for node in set(etsv.col_as_array_uniq("src") + etsv.col_as_array_uniq("target")):
        paths[node] = []
        all_paths[node] = []

    # create a map of children
    children_map = {}
    for (parent, cstr) in etsv.aggregate("src", ["target"], [funclib.uniq_mkstr]).to_tuples(["src", "target:uniq_mkstr"]):
        children = cstr.split(sep)
        children_map[parent] = children

    # create a map of parents
    parents_map = {}
    for (child, pstr) in etsv.aggregate("target", ["src"], [funclib.uniq_mkstr]).to_tuples(["target", "src:uniq_mkstr"]):
        parents = pstr.split(sep)
        parents_map[child] = parents

    # root nodes
    root_nodes = set(etsv.col_as_array_uniq("src")).difference(set(parents_map.keys()))

    # do a forward crawl to fill paths and all_paths
    forward_queue = queue.Queue()
    visited_parents = {}

    # add all roots
    for root in root_nodes:
        forward_queue.put(root)

    # while loop until queue is empty
    while (forward_queue.empty() == False):
        # remove an item
        node = forward_queue.get()

        # mark as visited
        visited_parents[node] = 1

        # get children
        children = children_map[node] if (node in children_map.keys()) else []

        # iterate and add
        for child in children:
            src = node
            dest = child

            all_paths[dest] = list(sorted(set(all_paths[dest] + all_paths[src] + [src])))
            if (dest not in paths[dest] and dest not in paths[src]):
                paths[dest] = list(sorted(set(paths[dest] + paths[src] + [src])))

            # append child to queue
            if (child not in visited_parents.keys()):
                forward_queue.put(child)

    # now do a backward run to include all paths to the parents
    for child in parents_map.keys():
        for parent in parents_map[child]:
            src = parent
            dest = child

            # get a combined list of all paths
            all_paths[dest] = list(sorted(set(all_paths[dest] + all_paths[src] + [src])))

    def __get_ancestor_paths__(x):
        results = []
        if (x in parents_map.keys()):
            for parent in parents_map[x]:
                if (x in paths.keys() and parent in paths[x]):
                    for ancestors in __get_ancestor_paths__(parent):
                        results.append([x] + ancestors)
        else:
           results.append([x])

        # return
        return results

    # find all unique paths to all the target nodes
    ancestors_map = {}
    for dest in etsv.col_as_array_uniq("target"):
        # do a backtrack and create unique paths
        ancestors = __get_ancestor_paths__(dest)
        ancestors_map[dest] = list([",".join(vs) for vs in ancestors])

    # return
    return etsv \
        .transform("target", lambda t: ",".join(paths[t]) if (t in paths.keys()) else "", "{}:src_paths".format(prefix)) \
        .transform("target", lambda t: ",".join(all_paths[t]) if (t in all_paths.keys()) else "", "{}:all_paths".format(prefix)) \
        .transform("target", lambda t: "|".join(ancestors_map[t]) if (t in ancestors_map.keys()) else "", "{}:ancestors".format(prefix))

def get_time_based_forward_edges_only(etsv, ts_col, prefix):
    utils.warn_once("get_time_based_forward_edges_only: this is hard to understand and time ordering is tricky. Use get_forward_edges_only")

    # list of ids found so far in crawl
    paths = {}
    all_paths = {} 

    # run loop
    sorted_edges = etsv \
        .filter(["src", "target"], lambda t1, t2: t1 != t2) \
        .numerical_sort([ts_col]) \
        .to_tuples(["src", "target"])

    # iterate over sorted edges
    for src, dest in sorted_edges:
        if (src not in all_paths.keys()):
            all_paths[src] = []

        if (dest not in all_paths.keys()):
            all_paths[dest] = []

        if (src not in paths.keys()):
            paths[src] = []

        if (dest not in paths.keys()):
            paths[dest] = []

        all_paths[dest] = list(sorted(set(all_paths[dest] + all_paths[src] + [src])))
        if (dest not in paths[dest] and dest not in paths[src]):
            paths[dest] = list(sorted(set(paths[dest] + paths[src] + [src])))

    # TODO: do a second pass to incorporate missing paths from first pass. this is not well understood.
    for src, dest in sorted_edges:
        all_paths[dest] = list(sorted(set(all_paths[dest] + all_paths[src] + [src])))

    # return
    return etsv \
        .transform("target", lambda t: ",".join(paths[t]) if (t in paths.keys()) else "", "{}:src_paths".format(prefix)) \
        .transform("target", lambda t: ",".join(all_paths[t]) if (t in all_paths.keys()) else "", "{}:all_paths".format(prefix))


def remove_dangling_edges(etsv, retain_vertex_ids, retain_node_filter_func, max_iter = 5, dmsg = ""):
    dmsg = utils.extend_inherit_message(dmsg, "remove_dangling_edges")
    utils.warn_once("{}: this can remove event retain keys if there is no incoming or outgoing edge".format(dmsg))

    # check for column names
    if (etsv.has_col("src") == False or etsv.has_col("target") == False):
        raise Exception("{}: predefined column names not found".format(dmsg))

    # flag to maintain current state
    dangling_edges_pruned = False
    count = 0

    # initialize results
    etsv_result = etsv

    # loop
    while (dangling_edges_pruned == False and count <= max_iter):
        # increase counter
        count = count + 1
    
        # count edges
        etsv_edge_count = etsv_result \
            .noop(["src", "target", "evports", "users", "ts_min", "ts_max", "count"], n = 1000, title = "etsv_result") \
            .aggregate(["target"], ["src"], [funclib.uniq_len], collapse = False) \
            .rename("src:uniq_len", "incoming_target") \
            .noop(["src", "target", "incoming_target"], n = 1000, title = "etsv_result incoming_target") \
            .aggregate(["src"], ["target"], [funclib.uniq_len], collapse = False) \
            .rename("target:uniq_len", "outgoing_target") \
            .noop(["src", "target", "outgoing_target"], n = 1000, title = "etsv_result outgoing_target") \
            .noop(["src", "target", "evports", "incoming_target", "outgoing_target"], n = 1000, title = "etsv_edge_count")
    
        # get outgoing edges
        etsv_num_outgoing_target = etsv_edge_count \
            .select(["src", "outgoing_target"]) \
            .rename("src", "right:src") \
            .distinct()
        
        # ege count flag
        etsv2_edge_flags = etsv_edge_count \
            .print_stats(msg = "etsv_edge_count") \
            .drop_cols(["outgoing_target"]) \
            .left_map_join(etsv_num_outgoing_target, ["target"], rkeys = ["right:src"], def_val_map = {"outgoing_target": "0"}) \
            .noop(["src", "target", "incoming_target", "outgoing_target", "right:.*"], n = 1000, title = "etsv_edge_count join") \
            .drop_cols_with_prefix("right") \
            .transform(["src", "target", "incoming_target", "outgoing_target"], lambda s,t,i,o: 1 if (retain_node_filter_func(s) and int(o) == 0) else 0, "outgoing_target_zero_flag") \
            .transform(["src", "target", "incoming_target", "outgoing_target"], lambda s,t,i,o: 1 if (retain_node_filter_func(s) and int(i) > 1) else 0, "incoming_target_mult_flag") \
            .sort(["outgoing_target_zero_flag", "incoming_target_mult_flag", "src", "target"]) \
            .noop(["src", "target", "incoming_target", "outgoing_target", "outgoing_target_zero_flag", "incoming_target_mult_flag"], n = 1000, title = "etsv2_edge_flags")

        etsv2_sync = etsv2_edge_flags \
            .noop(["src", "target", "outgoing_target_zero_flag", "incoming_target_mult_flag"], n = 1000, title = "etsv2_edge_flags 1") \
            .exclude_filter(["target", "outgoing_target_zero_flag"], lambda tgt, t: tgt not in retain_vertex_ids and t == "1") \
            .exclude_filter(["target", "incoming_target_mult_flag"], lambda tgt, t: tgt not in retain_vertex_ids and t == "1") \
            .exclude_filter(["target", "incoming_target_mult_flag", "outgoing_target_zero_flag"], lambda tgt, imulti, ozero: tgt in retain_vertex_ids and imulti == "1") \
            .noop(["src", "target", "outgoing_target_zero_flag", "incoming_target_mult_flag"], n = 1000, title = "etsv2_edge_flags 2") \
            .noop(["outgoing_target_zero_flag", "incoming_target_mult_flag"], lambda t1, t2: t1 == "1" or t2 == "1") \
            .noop(10000, "etsv2_sync", tsv.TSV.select, ["src", "target", "incoming_target", "outgoing_target", "data_source", "outgoing_target_zero_flag", "incoming_target_mult_flag"]) \
            .drop_cols(["incoming_target", "outgoing_target", "outgoing_target_zero_flag", "incoming_target_mult_flag"])

        # check for running the flag
        if (etsv2_edge_flags.num_rows() == etsv2_sync.num_rows()):
            dangling_edges_pruned = True
            utils.debug("etsv: no more dangling edges found")
        else:
            utils.info("etsv: dangling edges found. running the loop again : {} / {}".format(count, max_iter))

        # update the core data structure 
        etsv_result = etsv2_sync

    # return
    return etsv_result

# TODO: This method is a reference implementation
def remove_cycles(vtsv, etsv, ts_col, retain_node_filter_func = None, dmsg = ""):
    utils.extend_inherit_message(dmsg, "remove_cycles")
    utils.warn_once("remove_cycles: This logic needs to be corrected for same edges from multiple data_sources to not confuse each other")
    utils.warn_once("remove_cycles: there is a weird check for single edge. the t1 in t2.split is not clear")

    # validation
    if (vtsv.has_col("node_id") == False or etsv.has_col("src") == False or etsv.has_col("target") == False or etsv.has_col("data_source") == False or etsv.has_col(ts_col) == False):
        raise Exception("{}: predefined column names not found".format(dmsg))

    # edge count for the edges originating from the same data source 
    etsv_spl = etsv \
        .filter("src", retain_node_filter_func) \
        .aggregate(["src", "target", ts_col], ["data_source"], [funclib.uniq_mkstr]) \
        .distinct() \
        .noop(1000, title = "etsv_spl")

    # edge count for edges that are associated with vertices that need to be retained
    etsv_non_spl = etsv \
        .exclude_filter("src", retain_node_filter_func) \
        .aggregate(["src", "target", ts_col], ["data_source"], [funclib.uniq_mkstr]) \
        .distinct() \
        .noop(1000, title = "etsv_non_spl")
 
    # get forward edges and do some dedup based on ts_col 
    etsv_non_spl2 = get_time_based_forward_edges_only(etsv_non_spl, ts_col, "graph") \
        .sort([ts_col, "target", "src"]) \
        .transform("graph:src_paths", lambda t: ",".join([t1[0:4] for t1 in t.split(",")]) if (t != "") else "", "graph:src_paths2") \
        .transform("graph:all_paths", lambda t: ",".join([t1[0:4] for t1 in t.split(",")]) if (t != "") else "", "graph:all_paths2") \
        .noop(ts_col, funclib.utctimestamp_to_datetime_str, "graph:ts_min2") \
        .transform(["src", "graph:src_paths"], lambda t1, t2: 1 if (t1 in t2.split(",")) else 0, "graph:flag") \
        .noop(1000, "etsv_non_spl2 1", tsv.TSV.select, ["src", "target", ts_col, "graph:src_paths2", "graph:all_paths2", "graph:flag", "data_source:uniq_mkstr"]) \
        .eq_int("graph:flag", 1) \
        .drop_cols_with_prefix("graph") \
        .noop(1000, "etsv_non_spl2 2", tsv.TSV.select, ["src", "target", ts_col, "data_source:uniq_mkstr"])

    etsv2_included = tsv.merge_union([etsv_spl, etsv_non_spl2]) \
        .select(["src", "target", ts_col]) \
        .distinct() \
        .noop(1000, title = "etsv2_included") \
        .to_tuples(["src", "target", ts_col])

    # take only edges that have been picked to be included
    etsv2 = etsv \
        .filter(["src", "target", ts_col], lambda t1, t2, t3: (t1, t2, t3) in etsv2_included) \
        .sort(["src", "target"]) \
        .noop(1000, title = "etsv2", max_col_width = 20)

    # take only vertices that have edges
    vtsv2 = vtsv \
        .filter("node_id", lambda t: t in etsv2.col_as_array_uniq("src") + etsv2.col_as_array_uniq("target"))

    # return
    return vtsv2, etsv2

def merge_similar_nodes_reference(vtsv, etsv, retain_vertex_ids, ts_col, retain_node_filter_func, dmsg = ""):
    dmsg = utils.extend_inherit_message(dmsg, "merge_similar_nodes")

    # check for cycles
    vtsv_nocycle, etsv_nocycle = remove_cycles(vtsv, etsv, ts_col, retain_node_filter_func = retain_node_filter_func)
        
    # warn if cycles were present, and then remove them
    if (etsv.num_rows() != etsv_nocycle.num_rows()):
        utils.warn_once("merge_similar_nodes: this api is unpredictable in presence of cycles. removing them for robustness")
        # debug
        etsv \
            .filter(["src", "target"], lambda src, tgt: (src, tgt) not in etsv_nocycle.to_tuples(["src", "target"])) \
            .show_transpose(3, title = "{}: edges removed because of cycle".format(dmsg), max_col_width = 20)
        
        # assign to original variables
        vtsv, etsv = vtsv_nocycle, etsv_nocycle
        
    # remove columns that are created again
    etsv_sel = etsv

    # find the number of outgoing edges to mark leaves
    etsv_edge_count = etsv_sel \
        .aggregate(["target"], ["src"], [funclib.uniq_len], collapse = False) \
        .rename("src:uniq_len", "incoming_target") \
        .transform("target", lambda t: 1 if (t in retain_vertex_ids) else 0, "target_is_retain_vertex") \
        .transform(prop_col, lambda arr: ",".join(arr), "target_level2", use_array_notation = True) \
        .aggregate("src", ["target", "target_level2", "count"], [funclib.uniq_len, funclib.uniq_mkstr, funclib.sumint], collapse = False) \
        .rename("target:uniq_len", "outgoing_target") \
        .rename("target_level2:uniq_mkstr", "edge_target_level2") \
        .rename("count:sumint", "count_target_level2") \
        .drop_cols("target_level2")

    etsv_num_outgoing_target = etsv_edge_count \
        .select(["src", "outgoing_target", "edge_target_level2", "count_target_level2"]) \
        .rename("src", "right:src") \
        .distinct() \
        .noop("merge_similar_nodes: etsv_num_outgoing_target")

    etsv_target_edge_count = etsv_edge_count \
        .drop_cols(["outgoing_target", "edge_target_level2", "count_target_level2"]) \
        .left_map_join(etsv_num_outgoing_target, ["target"], rkeys = ["right:src"], def_val_map = {"outgoing_target": "0", "edge_target_level2": "", "count_target_level2": "0"}) \
        .drop_cols_with_prefix("right") \
        .reorder(["src", "target"])

    etsv_flags = etsv_target_edge_count \
        .transform(["src", "incoming_target", "outgoing_target"], lambda t1, t2, t3: 1 if (self.is_spl_node(t1) == False and (int(t2) <= 1 and int(t3) == 0)) else 0, "is_leaf_target") \
        .transform(["src", "incoming_target", "outgoing_target"], lambda t1, t2, t3: 1 if (int(t2) <= 1 and int(t3) == 1) else 0, "is_leaf_edge_target")

    etsv_flags \
        .exclude_filter(["is_leaf_target", "is_leaf_edge_target"], lambda t1, t2: t1 == "0" and t2 == "0")

    # group the leaf nodes based on collapse flag
    etsv_grouped_leaf0 = etsv_flags \
        .exclude_filter(["is_leaf_target", "is_leaf_edge_target"], lambda t1, t2: t1 == "1" or t2 == "1") \
        .add_const("num_nodes", "1") \
        .add_const("etsv_grouped_source", "leaf0")

    etsv_grouped_leaf1 = etsv_flags \
        .eq_str("is_leaf_target", "1") \
        .aggregate(["src", "target_is_retain_vertex", "is_leaf_target"], ["target", "target"], [funclib.uniq_mkstr, funclib.get_len]) \
        .rename("target:uniq_mkstr", "target") \
        .rename("target:get_len", "num_nodes") \
        .add_const("etsv_grouped_source", "leaf1")

    etsv_grouped_leaf2 = etsv_flags \
        .eq_str("is_leaf_edge_target", "1") \
        .aggregate(["src", "edge_target_level2", "is_leaf_edge_target"], ["target", "target"], [funclib.uniq_mkstr, funclib.get_len]) \
        .rename("target:uniq_mkstr", "target") \
        .rename("target:get_len", "num_nodes") \
        .add_const("etsv_grouped_source", "leaf2")

    etsv_grouped = tsv.merge_union([etsv_grouped_leaf0, etsv_grouped_leaf1, etsv_grouped_leaf2]) \
        .drop_cols(["incoming_target", "outgoing_target", "edge_target_level2", "count_target_level2", "is_leaf_target", "is_leaf_edge_target"]) \
        .reorder(["src", "target", "num_nodes"]) \
        .reverse_sort(["num_nodes"])

    # map of node ids for collapsed nodes
    vtsv_target_map = {}
    for t in etsv_grouped.col_as_array_uniq("target"):
        for t1 in t.split(","):
            vtsv_target_map[str(t1)] = str(t)

    # map for count of collapsed nodes         
    vtsv_node_count_map = {}
    for t1, t2 in etsv_grouped.to_tuples(["target", "num_nodes"]):
        vtsv_node_count_map[str(t1)] = str(t2)

    # create summarized columns. TODO: this reference is confusing
    etsv_grouped2 = etsv_grouped \
        .transform_inline("src", lambda t: vtsv_target_map[t] if (t in vtsv_target_map.keys()) else t)

    # create new ids. TODO: poor reference
    vtsv_grouped = vtsv \
        .transform_inline("node_id", lambda t: vtsv_target_map[t] if (t in vtsv_target_map.keys()) else t) \
        .aggregate(["node_id"], ["__is_root__", "__is_retain_vertex__"],
            [funclib.uniq_mkstr, funclib.uniq_mkstr]) \
        .remove_suffix("uniq_mkstr", dmsg = dmsg) \
        .transform("node_id", lambda t: vtsv_node_count_map[t] if (t in vtsv_node_count_map.keys()) else "0", "num_nodes")

    # return 
    return vtsv_grouped, etsv_grouped2

# retain_vertex_annotations are the start and end timestamps for the retained vertex ids
def split_graph_filter_func(src, tgt, ts, retain_vertex_ids, retain_vertex_annotations, retain_node_filter_func, dmsg = ""):
    dmsg = utils.extend_inherit_message(dmsg, "split_graph_filter_func")

    # return True for special nodes
    if (retain_node_filter_func(src) or retain_node_filter_func(tgt)):
        return True

    # return True if neither are detections
    if (src not in retain_vertex_ids and tgt not in retain_vertex_ids):
        return True

    # check before and after flag
    before_flag = True
    after_flag = True

    # important, for both nodes as detection, keep the edge
    if (src in retain_vertex_ids and tgt in retain_vertex_ids):
        return True

    # before detection
    if (tgt in retain_vertex_ids):
        if (tgt in retain_vertex_annotations.keys()):
            (retain_vertex_ts_min, retain_vertex_ts_max) = retain_vertex_annotations[tgt]
            if (int(ts) > int(retain_vertex_ts_max)):
                before_flag = False

    # after detection
    if (src in retain_vertex_ids):
        if (src in retain_vertex_annotations.keys()):
            (retain_vertex_ts_min, retain_vertex_ts_max) = retain_vertex_annotations[src]
            if (int(retain_vertex_ts_min) > int(ts)):
                after_flag = False

    # default
    if (before_flag == False or after_flag == False):
        return False
    else:
        return True

def apply_time_order_based_filter(vtsv, etsv, retain_vertex_ids, retain_node_filter_func, strict_ordering_flag):
    # find the min and max timestamps
    etsv_min_max = etsv \
        .aggregate(["src", "target"], ["ts", "ts"], [funclib.minint, funclib.maxint]) \
        .rename("ts:minint", "ts_min") \
        .rename("ts:maxint", "ts_max")

    # create left and right sides. apply timestamp ordering logic to the subgraph to further prune edges
    etsv_left = etsv_min_max.select(["src", "target", "ts_min"]).exclude_filter("src", retain_node_filter_func).add_prefix("left").distinct()
    etsv_right = etsv_min_max.select(["src", "target", "ts_max"]).exclude_filter("src", retain_node_filter_func).add_prefix("right").distinct()

    def __apply_time_order_based_filter_template__(lflag, rflag, ldetect, rdetect):
        # if strict time ordering is asked, remove the forward edge
        if (strict_ordering_flag == True):
            return "right"

        # check for detect flags
        if (ldetect == 1 and rdetect == 1):
            return ""
        elif (ldetect == 1):
            return "right"
        elif (rdetect == 1):
            return "left"
        elif (lflag == "0"):
            return "left"
        elif (rflag == "0"):
            return "right"
        else:
            raise Exception("Invalid parameters: lflag: {}, rflag: {}, ldetect: {}, rdetect: {}".format(lflag, rflag, ldetect, rdetect))

    # excluded edges
    excluded_edges = etsv_left \
        .inner_map_join(etsv_right, ["left:target"], rkeys = ["right:src"]) \
        .transform(["left:ts_min", "right:ts_max"], lambda t1,t2: 1 if (int(t1) <= int(t2)) else 0, "ts:flag") \
        .transform_inline(["left:ts_min", "right:ts_max"], funclib.utctimestamp_to_datetime_str) \
        .reorder(["left:src", "left:target", "right:src", "right:target"], use_existing_order = False) \
        .aggregate(["left:src", "left:target"], ["ts:flag"], [funclib.uniq_mkstr], collapse = False) \
        .rename("ts:flag:uniq_mkstr", "left:ts:flag:uniq_mkstr") \
        .aggregate(["right:src", "right:target"], ["ts:flag"], [funclib.uniq_mkstr], collapse = False) \
        .rename("ts:flag:uniq_mkstr", "right:ts:flag:uniq_mkstr") \
        .filter(["left:ts:flag:uniq_mkstr", "right:ts:flag:uniq_mkstr"], lambda t1, t2: t1 == "0" or t2 == "0") \
        .select(["left:src", "left:target", "right:src", "right:target", "left:ts:flag:uniq_mkstr", "right:ts:flag:uniq_mkstr"]) \
        .distinct() \
        .transform(["left:src", "left:target"], lambda t1, t2: 1 if (t1 in retain_vertex_ids or t2 in retain_vertex_ids) else 0, "left:is_detect") \
        .transform(["right:src", "right:target"], lambda t1, t2: 1 if (t1 in retain_vertex_ids or t2 in retain_vertex_ids) else 0, "right:is_detect") \
        .transform(["left:ts:flag:uniq_mkstr", "right:ts:flag:uniq_mkstr", "left:is_detect", "right:is_detect"], lambda lflag, rflag, ldetect, rdetect:
            __apply_time_order_based_filter_determine_side__(strict_time_ordering_flag, lflag, rflag, int(ldetect), int(rdetect)), "exclude_side") \
        .is_nonempty_str("exclude_side") \
        .transform(["left:src", "left:target", "right:src", "right:target", "exclude_side"], lambda lsrc, ltgt, rsrc, rtgt, eside:
            (lsrc, ltgt) if (eside == "left") else (rsrc, rtgt), ["excluded:src", "excluded:target"]) \
        .select(["excluded:src", "excluded:target"]) \
        .distinct()

    # apply exclusion
    etsv_result = etsv \
        .exclude_filter(["src", "target"], lambda src, tgt: (src, tgt) in excluded_edges.to_tuples(["excluded:src", "excluded:target"]))
    vtsv_result = vtsv \
        .values_in("node_id", lambda t: t in etsv_result.col_as_array_uniq("src") + etsv_result.col_as_array_uniq("target"))

    # return
    return vtsv_result, etsv_result




"""library of function to be used in calling lambda functions"""

import statistics
import numpy as np
from dateutil import parser
import datetime
from omigo_core import utils

def file_base_name(x):
    if (len(x) <= 1):
        return x

    index = -1
    if ("\\" in x):
        index = x.rindex("\\")
    elif ("/" in x):
        index = x.rindex("/")
    return str(x[index+1:])

def get_len(vs):
    return str(len(vs))

def get_non_empty_len(vs):
    vs = list(filter(lambda t: len(t.strip()) > 0, vs))
    return str(len(vs))

def uniq_len(vs):
    vs2 = set()
    for t in vs:
        for k in str(t).split(","):
            if (len(k.strip()) > 0):
                vs2.add(str(k))
    return str(len(vs2))

def uniq_mkstr(vs):
    vs2 = set()
    for t in vs:
        for k in str(t).split(","):
            if (len(k.strip()) > 0):
                vs2.add(str(k))
    return ",".join(sorted([str(x) for x in vs2]))

def split_merge_uniq_mkstr(vs):
    vs2 = []
    for v in vs:
        vs2 = vs2 + v.split(",")
    vs2 = list(set(filter(lambda t: len(t.strip()) > 0, vs2)))
    return ",".join(sorted([str(x) for x in vs2]))

def mean(vs):
    vs = list([float(v) for v in vs])
    return statistics.mean(vs)

def std_dev(vs):
    vs = list([float(v) for v in vs])

    # adding check for atleast 2 elements before calling the statistics package
    if (len(vs) < 2):
        return 0

    # return
    return statistics.stdev(vs)

def mkstr(vs):
    vs2 = list(filter(lambda t: len(t.strip()) > 0, [str(x) for x in vs]))
    return ",".join(vs2)

def sorted_mkstr(vs):
    vs2 = sorted(list(filter(lambda t: len(t.strip()) > 0, [str(x) for x in vs])))
    return ",".join(vs2)

def mkstr4f(vs):
    vs2 = list(["{:4f}".format(float(x)) for x in vs])
    return ",".join(vs2)

def minint(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("minint: empty vs")

    min_value = str(vs[0])
    for v in vs[1:]:
        if (int(float(v)) < int(float(min_value))):
            min_value = str(v)

    return str(min_value)

def maxint(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("maxint: empty vs")

    max_value = str(vs[0])
    for v in vs[1:]:
        if (int(float(v)) > int(float(max_value))):
            max_value = str(v)

    return str(max_value)

def minfloat(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("minfloat: empty vs")

    min_value = str(vs[0])
    for v in vs[1:]:
        if (float(v) < float(min_value)):
            min_value = str(v)

    return str(min_value)

def maxfloat(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("maxfloat: empty vs")

    max_value = str(vs[0])
    for v in vs[1:]:
        if (float(v) > float(max_value)):
            max_value = str(v)

    return str(max_value)

def minstr(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("minstr: empty vs")

    min_value = str(vs[0])
    for v in vs[1:]:
        if (str(v) < min_value):
            min_value = str(v)

    return str(min_value)

def maxstr(vs):
    if (vs is None or len(vs) == 0):
        raise Exception("maxstr: empty vs")

    max_value = str(vs[0])
    for v in vs[1:]:
        if (str(v) > max_value):
            max_value = str(v)

    return str(max_value)

def minint_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return ""
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return ""
        else:
            return minint(vs)

def maxint_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return ""
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return ""
        else:
            return maxint(vs)

def minstr_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return ""
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return ""
        else:
            return minstr(vs)

def maxstr_failsafe(vs):
    if (vs is None or len(vs) == 0):
        return ""
    else:
        vs = list(filter(lambda t: t != "", vs))
        if (len(vs) == 0):
            return ""
        else:
            return maxstr(vs)

def sumint(vs):
    if (len(vs) == 0):
        return 0
    else:
        return sum([int(float(t)) for t in vs])

def sumfloat(vs):
    if (len(vs) == 0):
        return 0.0
    else:
        return sum([float(t) for t in vs])

# TODO: The semantics are not clear
def uniq_count(vs):
    vs2 = list(filter(lambda t: t.strip() != "", vs))
    return len(set(vs2))

def merge_uniq(vs):
    result = []
    for v in vs:
        for v2 in v.split(","):
            result.append(v2)

    # return
    return ",".join(sorted(list(set(result))))

def select_first(x, y):
    return x

def select_max_int(x, y):
    return int(max(int(x), int(y)))

def quantile(xs, start = 0, end = 1, by = 0.25, precision = 4):
    if (start > end):
        raise Exception("Start: {} > End: {}".format(start, end))

    qarr = []
    cur = start
    while (cur < end):
        qarr.append(cur)
        cur = cur + by

    format_str = "{:." + str(precision) + "f}"
    quan = np.quantile(str_arr_to_float(xs), qarr)
    return ",".join(format_str.format(x) for x in quan)

def quantile4(xs):
    return quantile(xs)

def quantile10(xs):
    return quantile(xs, by=1/10)

def quantile40(xs):
    return quantile(xs, by=1/40)

def max_str(xs):
    utils.warn_once("max_str is deprecated. Use maxstr")
    xs = sorted(xs)
    return xs[-1]

def min_str(xs):
    utils.warn_once("min_str is deprecated. Use minstr")
    xs = sorted(xs)
    return xs[0]

def to2digit(x):
    return "{:.2f}".format(float(x))

def to4digit(x):
    return "{:.4f}".format(float(x))

def to6digit(x):
    return "{:.6f}".format(float(x))

def convert_prob_to_binary(x, split=0.5):
    if (x >= split):
        return 1
    else:
        return 0

def select_first_non_empty(*args, **kwargs):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        return ""

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # boundary conditions
    for x in xs:
        if (x is not None and x != ""):
            return str(x)

    # return default
    return ""

# def if_else_non_empty_str(x, v1, v2):
def if_else_non_empty_str(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_non_empty_str: invalid input")

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 3):
        raise Exception("if_else_non_empty_str: invalid input: {}".format(xs))

    # assign variables
    x = xs[0]
    v1 = xs[1]
    v2 = xs[2]

    if (x is not None and str(x) != ""):
        return str(v1)
    else:
        return str(v2)

# def if_else_str(xinput, xval, v1, v2):
def if_else_str(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_non_empty_str: invalid input")

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 4):
        raise Exception("if_else_str: invalid input: {}".format(xs))

    # assign variables
    xinput = xs[0]
    xval = xs[1]
    v1 = xs[2]
    v2 = xs[3]

    # apply function
    if (xinput is not None and str(xinput) == str(xval)):
        return str(v1)
    else:
        return str(v2)

# def if_else_int(xinput, xval, v1, v2):
def if_else_int(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_int: invalid input")

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 4):
        raise Exception("if_else_int: invalid input: {}".format(xs))

    # assign variables
    xinput = xs[0]
    xval = xs[1]
    v1 = xs[2]
    v2 = xs[3]

    # apply function
    if (xinput is not None and int(xinput) == int(xval)):
        return v1
    else:
        return v2

# def if_else_non_zero_int(x, v1, v2):
def if_else_non_zero_int(*args):
    # variable name
    xs = args

    # boundary condition
    if (xs is None or len(xs) == 0):
        raise Exception("if_else_non_zero_int: invalid input")

    # check if it is list or tuple
    if (isinstance(xs[0], list)):
        xs = xs[0]

    # validation
    if (len(xs) != 3):
        raise Exception("if_else_int: invalid input: {}".format(xs))

    # assign variables
    x = xs[0]
    v1 = xs[1]
    v2 = xs[2]

    # apply function
    if (x is None or int(x) != 0):
        return v1
    else:
        return v2




import urllib
import json
import requests
from omigo_core import tsv, utils

# https://stackoverflow.com/questions/29931671/making-an-api-call-in-python-with-an-api-that-requires-a-bearer-token
class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token
    def __call__(self, r):
        r.headers["Authorization"] = "Bearer " + self.token
        return r

# TODO: the body has to be a json payload. This is because of some bug in python requests.post api
# TODO: allow_redirects=False
# TODO: https://docs.python-requests.org/en/latest/user/quickstart/: Check the WARNINGS
def __read_base_url__(url, query_params = {}, headers = {}, body = None, username = None, password = None, api_token = None, timeout_sec = 120, verify = True, method = None, dmsg = ""):
    dmsg = utils.extend_inherit_message(dmsg, "__read_base_url__")

    # check for query params
    if (len(query_params) > 0):
        params_encoded_str = urllib.parse.urlencode(query_params)
        url = "{}?{}".format(url, params_encoded_str)

    # determine method
    if (method is None):
        if (body is None):
            method = "GET"
        else:
            method = "POST"

    # exception handling
    try:
        # call the web service
        if (method == "GET"):
            # make web service call
            if (username is not None and password is not None):
                response = requests.get(url, auth = (username, password), headers = headers, timeout = timeout_sec, verify = verify)
            elif (api_token is not None):
                response = requests.get(url, auth = BearerAuth(api_token), headers = headers, timeout = timeout_sec, verify = verify)
            else:
                response = requests.get(url, headers = headers, timeout = timeout_sec, verify = verify)
        elif (method == "POST"):
            # do some validation on body
            body_json = None
            try:
                body_json = json.loads(body)
            except Exception as e:
                utils.error("{}: body is not well formed json: {}".format(dmsg, body))
                raise e

            # make web service call
            if (username is not None and password is not None):
                response = requests.post(url, auth = (username, password), json = body_json, headers = headers, timeout = timeout_sec, verify = verify)
            elif (api_token is not None):
                response = requests.post(url, auth = BearerAuth(api_token), json = body_json, headers = headers, timeout = timeout_sec, verify = verify)
            else:
                response = requests.post(url, json = body_json, headers = headers, timeout = timeout_sec, verify = verify)
        elif (method == "PUT"):
            # do some validation on body
            body_json = None
            try:
                body_json = json.loads(body)
            except Exception as e:
                utils.error("{}: body is not well formed json: {}".format(dmsg, body))
                raise e

            # make web service call
            if (username is not None and password is not None):
                response = requests.put(url, auth = (username, password), json = body_json, headers = headers, timeout = timeout_sec, verify = verify)
            elif (api_token is not None):
                response = requests.put(url, auth = BearerAuth(api_token), json = body_json, headers = headers, timeout = timeout_sec, verify = verify)
            else:
                response = requests.put(url, json = body_json, headers = headers, timeout = timeout_sec, verify = verify)
        elif (method == "DELETE"):
            if (username is not None and password is not None):
                response = requests.delete(url, auth = (username, password), headers = headers, timeout = timeout_sec, verify = verify)
            elif (api_token is not None):
                response = requests.delete(url, auth = BearerAuth(api_token), headers = headers, timeout = timeout_sec, verify = verify)
            else:
                response = requests.delete(url, headers = headers, timeout = timeout_sec, verify = verify)
        else:
            raise Exception("Invalid method: {}".format(method))

        # return response
        return response, None
    except requests.exceptions.RequestException as e:
        utils.warn("{}: Found exception while making request: {}".format(dmsg, e))
        return None, e

# TODO: the semantics of this api are not clear
def read_url_json(url, query_params = {}, headers = {}, body = None, username = None, password = None, api_token = None, timeout_sec = 120, verify = True, method = None):
    utils.warn("read_url_json will flatten json that comes out as list. This api is still under development")

    # read response
    response_str, status_code, error_msg = read_url_response(url, query_params, headers, body = body, username = username, password = password, api_token = api_token,
        timeout_sec = timeout_sec, verify = verify, method = method)

    # construct header
    header = "\t".join(["json_encoded", "status_code", "error_msg"])
    data = []

    # look for error conditions
    if (status_code == 200):
        # parse json object
        json_obj = json.loads(response_str)

        # based on the type of the response, flatten the structure
        if (isinstance(json_obj, list)):
            # iterate and add as row
            for v in json_obj:
                fields = [utils.url_encode(json.dumps(v).replace("\n", " ")), str(status_code), str(error_msg)]
                data.append("\t".join(fields))
        elif (isinstance(json_obj, dict)):
            fields = [utils.url_encode(json.dumps(json_obj).replace("\n", " ")), str(status_code), str(error_msg)]
            data.append("\t".join(fields))
        else:
            fields = ["", "0", "Unable to parse the json response: {}".format(response_str)]
            data.append("\t".join(fields))
    else:
        fields = ["", "0", "Unable to parse the json response: {}".format(utils.url_encode(response_str).replace("\n", " "))]
        data.append("\t".join(fields))

    return tsv.TSV(header, data).validate()

def read_url_response(url, query_params = {}, headers = {}, body = None, username = None, password = None, api_token = None, timeout_sec = 120, verify = True, method = None,
    num_retries = 1, retry_sleep_sec = 1):
    # read response
    response, resp_exception = __read_base_url__(url, query_params, headers, body = body, username = username, password = password, api_token = api_token,
        timeout_sec = timeout_sec, verify = verify, method = method)

    # check for errors
    if (response is None):
        return "", 500, str(resp_exception)

    # check for error codes
    if (response.status_code != 200):
        # check for website saying too many requests (429) or service unavailable (503)
        if (response.status_code == 429 or response.status_code == 503):
            # too many requests. wait and try again.
            if (num_retries > 0):
                utils.debug("read_url_response: url: {}, query_params: {}, got status: {}. Attempts remaining: {}. Retrying after sleeping for {} seconds".format(url, query_params,
                    response.status_code, num_retries, retry_sleep_sec))
                time.sleep(retry_sleep_sec)
                return read_url_response(url, query_params = query_params, headers = headers, body = body, username = username, password = password, api_token = api_token,
                    timeout_sec = timeout_sec, verify = verify, method = method, num_retries = num_retries - 1, retry_sleep_sec = retry_sleep_sec * 2)
            else:
                utils.debug("read_url_response: url: {}, getting 429 too many requests or 5XX error. Use num_retries parameter to for backoff and retry.".format(url))
                return "", response.status_code, response.reason
        else:
            return "", response.status_code, response.reason

    # check for content type
    content_type = None
    if ("content-type" in response.headers):
        content_type = response.headers["content-type"].lower()
    elif ("Content-Type" in response.headers):
        content_type = response.headers["Content-Type"].lower()
    else:
        return "", 0, "Missing content-type: {}".format(str(response.headers))

    # best guess file type
    file_type = None
    if (url.endswith(".csv") or url.endswith(".tsv")):
        file_type = "text"
    elif (url.endswith(".gz")):
        file_type = "gzip"
    elif (url.endswith(".zip")):
        file_type = "zip"

    # get response
    response_str = None
    if (content_type.startswith("application/x-gzip-compressed") or content_type.startswith("application/gzip") or file_type == "gzip"):
        response_str = str(gzip.decompress(response.content), "utf-8").rstrip("\n")
    elif (content_type.startswith("application/x-zip-compressed") or content_type.startswith("application/zip") or file_type == "zip"):
        barr = response.content
        zfile = zipfile.ZipFile(BytesIO(barr))
        response_str = zfile.open(zfile.infolist()[0]).read().decode().rstrip("\n")
        zfile.close()
    elif (content_type.startswith("text/plain") or content_type.startswith("application/json") or file_type == "text"):
        response_str = response.content.decode().rstrip("\n")
    elif (content_type.startswith("application/octet-stream")):
        utils.warn("Content Type is octet stream. Using gzip.".format(content_type))
        response_str = str(gzip.decompress(response.content), "utf-8").rstrip("\n")
    else:
        utils.warn("Content Type is not known: {}. Using default decoding from str.decode().".format(content_type))
        response_str = response.content.decode().rstrip("\n")

    # return
    return response_str, response.status_code, ""

# TODO: Deprecated
def read_url(url, query_params = {}, headers = {}, sep = None, username = None, password = None, api_token = None, timeout_sec = 120, verify = True, method = None):
    utils.warn("This method name is deprecated. Use read_url_as_tsv instead")
    return read_url_as_tsv(url, query_params = query_params, headers = headers, sep = sep, username = username, password = password, api_token = api_token,
    timeout_sec = timeout_sec, verify = verify, method = method)

# TODO: the compressed file handling should be done separately in a function
def read_url_as_tsv(url, query_params = {}, headers = {}, sep = None, username = None, password = None, api_token = None, timeout_sec = 120, verify = True, method = None):
    # use the file extension as alternate way of detecting type of file
    # TODO: move the file type and extension detection to separate function
    file_type = None
    if (url.endswith(".csv") or url.endswith(".tsv")):
        file_type = "text"
    elif (url.endswith(".gz")):
        file_type = "gzip"
    elif (url.endswith(".zip")):
        file_type = "zip"

    # detect extension
    ext_type = None
    if (url.endswith(".csv") or url.endswith(".csv.gz") or url.endswith(".csv.zip")):
        ext_type = "csv"
        utils.warn("CSV file detected. Only simple csv files are supported where comma and tabs are not part of any data.")
    elif (url.endswith(".tsv") or url.endswith(".tsv.gz") or url.endswith(".tsv.zip")):
        ext_type = "tsv"
    else:
        utils.warn("Unknown file extension. Doing best effort in content type detection")

    # read response
    response_str, status_code, error_msg, = read_url_response(url, query_params, headers, body = None, username = username, password = password, api_token = api_token,
        timeout_sec = timeout_sec, verify = verify, method = method)

    # check for status code
    if (status_code != 200):
        raise Exception("read_url failed. Status: {}, Reason: {}".format(status_code, error_msg))

    # split into lines
    lines = response_str.split("\n")
    header = lines[0]
    data = lines[1:]

    # check for separator
    if (sep is None and "\t" not in response_str):
        if ("," in header or ext_type == "csv"):
            sep = ","

    # check for other separators
    if (sep is not None and sep != "\t"):
        # do a better version of separator detection
        if ("\t" in response_str):
            utils.warn("Non TSV input file has tabs. Converting tabs to spaces")
            header = header.replace("\t", " ")
            data = [x.replace("\t", " ") for x in data]

        # replace separator with tabs
        header = header.replace(sep, "\t")
        data = [x.replace(sep, "\t") for x in data]

    # return
    return tsv.TSV(header, data).validate()




"""library of function to be used in general purpose data processing for TSV class"""

import statistics
import numpy as np
from dateutil import parser
import datetime
from omigo_core import utils
from omigo_core import udfs, timefuncs

# TODO: mkstr variants needs to use *args

def parse_image_file_base_name(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.parse_file_base_name(*args, **kwargs)

def get_len(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.get_len(*args, **kwargs)

def get_non_empty_len(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.get_non_empty_len(*args, **kwargs)

def uniq_len(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.uniq_len(*args, **kwargs)

def uniq_mkstr(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.uniq_mkstr(*args, **kwargs)

def split_merge_uniq_mkstr(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.split_merge_uniq_mkstr(*args, **kwargs)

def mean(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.mean(*args, **kwargs)

def std_dev(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.std_dev(*args, **kwargs)

def mkstr(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.mkstr(*args, **kwargs)

def sorted_mkstr(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.sorted_mkstr(*args, **kwargs)

def mkstr4f(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.mkstr4f(*args, **kwargs)

def minint(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.minint(*args, **kwargs)

def maxint(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.maxint(*args, **kwargs)

def minfloat(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.minfloat(*args, **kwargs)

def maxfloat(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.maxfloat(*args, **kwargs)

def minstr(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.minstr(*args, **kwargs)

def maxstr(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.maxstr(*args, **kwargs)

def minint_failsafe(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.minint_failsafe(*args, **kwargs)

def maxint_failsafe(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.maxint_failsafe(*args, **kwargs)

def minstr_failsafe(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.minstr_failsafe(*args, **kwargs)
            
def maxstr_failsafe(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.maxstr_failsafe(*args, **kwargs)

def sumint(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.sumint(*args, **kwargs)

def sumfloat(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.sumfloat(*args, **kwargs)

# TODO: The semantics are not clear
def uniq_count(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.uniq_count(*args, **kwargs)

def merge_uniq(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.merge_uniq(*args, **kwargs)

def select_first(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.select_first(*args, **kwargs)

def select_first_non_empty(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.select_first_non_empty(*args, **kwargs)

def select_max_int(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.select_max_int(*args, **kwargs)

def str_arr_to_float(xs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return [float(x) for x in xs]

def quantile(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.quantile(*args, **kwargs)

def quantile4(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.quantile4(*args, **kwargs)

def quantile10(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.quantile10(*args, **kwargs)

def quantile40(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.quantile40(*args, **kwargs)

def max_str(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.max_str(*args, **kwargs)

def min_str(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.min_str(*args, **kwargs)

def to2digit(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.to2digit(*args, **kwargs)

def to4digit(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.to2digit(*args, **kwargs)

def to6digit(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.to2digit(*args, **kwargs)

def convert_prob_to_binary(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use udfs package")
    return udfs.convert_prob_to_binary(*args, **kwargs)

def get_str_map_with_keys(mp, keys, fail_on_missing = True):
    mp2 = {}
    for k in keys:
        if (k in mp.keys()):
            mp2[k] = str(mp[k])
        else:
            if (fail_on_missing):
                raise Exception("Missing key: {}, avaialble keys: {}".format(k, str(mp.keys())))

    return mp2

def get_str_map_without_keys(mp, excluded_keys):
    mp2 = {}
    for k in mp.keys():
        if (k not in excluded_keys):
            mp2[k] = str(mp[k])

    return mp2

# def if_else_non_empty_str(x, v1, v2):
def if_else_non_empty_str(*args):
    return udfs.if_else_non_empty_str(*args)

# def if_else_str(xinput, xval, v1, v2):
def if_else_str(*args):
    return udfs.if_else_str(*args)

# def if_else_int(xinput, xval, v1, v2):
def if_else_int(*args):
    return udfs.if_else_int(*args)

# def if_else_non_zero_int(x, v1, v2):
def if_else_non_zero_int(*args):
    return udfs.if_else_non_zero_int(*args)

# TODO: this is bad implementation. The Win32 timestamp format needs proper handling
def win32_timestamp_to_utctimestamp(x):
    return int(str(x)[0:-8]) + 339576461

def get_time_diffs(vs):
    # sort the input
    vs = sorted(list([datetime_to_utctimestamp_sec(t) for t in vs]))

    # boundary condition
    if (len(vs) <= 1):
        return ""

    # get parirs
    pairs = list(zip(vs[0:len(vs)-1], vs[1:]))

    # result
    result = []
    for pair in pairs:
        v1, v2 = pair
        diff = int(v2 - v1)
        if (diff < 60):
            result.append("{}s".format(diff))
        elif (diff < 60 * 60):
            result.append("{}m".format(int(diff / 60)))
        elif (diff < 24 * 60 * 60):
            result.append("{}h".format(int(diff / (60 * 60))))
        elif (diff < 30 * 24 * 60 * 60):
            result.append("{}d".format(int(diff / (24 * 60 * 60))))
        else:
            result.append("{}:s".format(diff))

    # return
    return ",".join(result)

def simple_map_to_url_encoded_col_names(cols, url_encoded_cols = None):
    # create result
    results = []
    if (url_encoded_cols is not None):
        # iterate
        for c in cols:
            if (c in url_encoded_cols):
                results.append("{}:url_encoded".format(c))
            else:
                results.append(c)
    else:
        results = cols

    # return
    return results

def map_to_url_encoded_col_names(cols, prefix = None, url_encoded_cols = None):
    results = []

    # iterate
    for c in cols:
        col = c
        # check if there is prefix 
        if (col.find(":") != -1):
            col = col.split(":")[-1]

        # assign
        result = c

        # handle url_encoded suffix
        if (url_encoded_cols is not None and col in url_encoded_cols):
            result = "{}:url_encoded".format(c)

        # handle prefix
        if (prefix is not None):
            result = "{}:{}".format(prefix, result)

        # append
        results.append(result)

    # return
    return results

def get_display_relative_time_str(v):
    # convert to int
    v = int(v)

    # compute units
    days = v // 86400
    hours = (v - (days * 86400)) // 3600
    minutes = (v - (days * 86400 + hours * 3600)) // 60
    seconds = v - (days * 86400 + hours * 3600 + minutes * 60)
            
    # compute display string
    results = []
    count = 0
    max_display_values = 2
    if (days > 0 and count < max_display_values):
        results.append("{}d".format(days))
        count = count + 1
    if (hours > 0 and count < max_display_values):
        results.append("{}h".format(hours))
        count = count + 1
    if (minutes > 0 and count < max_display_values):
        results.append("{}m".format(minutes))
        count = count + 1
    if (seconds > 0 and count < max_display_values):
        results.append("{}s".format(seconds))
        count = count + 1

    # get string
    result = " ".join(results)

    # return
    return result

def datetime_to_utctimestamp_millis(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.datetime_to_utctimestamp_millis(*args, **kwargs)

def datetime_to_utctimestamp(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.datetime_to_utctimestamp(*args, **kwargs)

def datetime_to_utctimestamp_sec(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.datetime_to_utctimestamp_sec(*args, **kwargs)

def utctimestamp_to_datetime(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.utctimestamp_to_datetime(*args, **kwargs)

def utctimestamp_millis_to_datetime(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.utctimestamp_millis_to_datetime(*args, **kwargs)

def utctimestamp_to_datetime_str(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.utctimestamp_to_datetime_str(*args, **kwargs)

def utctimestamp_millis_to_datetime_str(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.utctimestamp_millis_to_datetime_str(*args, **kwargs)

def datetime_to_timestamp(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.datetime_to_timestamp(*args, **kwargs)

def get_utctimestamp_sec(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.get_utctimestamp_sec(*args, **kwargs)

def get_utctimestamp_millis(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return timefuncs.get_utctimestamp_millis(*args, **kwargs)

def datestr_to_datetime(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use timefuncs package")
    return datestr_to_datetime(*args, **kwargs)




"""utlity methods to read and write tsv data."""

from urllib.parse import urlencode
#from urllib.request import Request, urlopen
#from urllib.error import HTTPError, URLError
from io import BytesIO
import gzip
import json
import time
import zipfile
import requests

# local imports
from omigo_core import tsv
from omigo_core import file_paths_util
from omigo_core import file_paths_data_reader
from omigo_core import file_io_wrapper
from omigo_core import utils
from requests import exceptions
from omigo_hydra import hydra

# TODO: find the difference between ascii and utf-8 encoding
# requests.post doesnt take data properly. Use json parameter.
# TODO: use the local_fs_wrapper to use shared code for reading and writing

# TODO: need to document that a simple union can be achieved by setting def_val_map = {}
def merge(tsv_list, def_val_map = None):
    # validation
    if (len(tsv_list) == 0):
        utils.warn("Error in input. List of tsv is empty")
        return tsv.create_empty()

    # remove tsvs without any columns
    tsv_list = list(filter(lambda x: x.num_cols() > 0, tsv_list))

    # base condition
    if (len(tsv_list) == 0):
        utils.warn("List of tsv is empty. Returning")
        return tsv.create_empty()

    # warn if a huge tsv is found 
    for i in range(len(tsv_list)):
        if (tsv_list[i].size_in_gb() >= 1):
            utils.warn("merge: Found a very big tsv: {} / {}, num_rows: {}, size (GB): {}. max_size_cols_stats: {}".format(
                i + 1, len(tsv_list), tsv_list[i].num_rows(), tsv_list[i].size_in_gb(), str(tsv_list[i].get_max_size_cols_stats())))
            tsv_list[i].show_transpose(1, title = "merge: big tsv")

    # check for valid headers
    header = tsv_list[0].get_header()
    header_fields = tsv_list[0].get_header_fields()

    # iterate to check mismatch in header
    index = 0
    for t in tsv_list:
        # Use a different method for merging if the header is different
        if (header != t.get_header()):
            header_diffs = get_diffs_in_headers(tsv_list)

            # display warning according to kind of differences found
            if (len(header_diffs) > 0):
                # TODO
                if (def_val_map is None):
                    utils.warn("Mismatch in header at index: {}. Cant merge. Using merge_intersect for common intersection. Some of the differences in header: {}".format(
                        index, str(header_diffs)))
            else:
                utils.warn("Mismatch in order of header fields: {}, {}. Using merge intersect".format(header.split("\t"), t.get_header().split("\t")))

            # return
            return merge_intersect(tsv_list, def_val_map = def_val_map)

        # increment
        index = index + 1

    # simple condition
    if (len(tsv_list) == 1):
        return tsv_list[0]
    else:
        return tsv_list[0].union(tsv_list[1:])

def split_headers_in_common_and_diff(tsv_list):
    common = {}
    # get the counts for each header field
    for t in tsv_list:
        for h in t.get_header_fields():
            if (h not in common.keys()):
                common[h] = 0
            common[h] = common[h] + 1

    # find the columns which are not present everywhere
    non_common = []
    for k, v in common.items():
        if (v != len(tsv_list)):
            non_common.append(k)

    # return
    return sorted(common.keys()), sorted(non_common)

def get_diffs_in_headers(tsv_list):
    common, non_common = split_headers_in_common_and_diff(tsv_list)
    return non_common

def merge_intersect(tsv_list, def_val_map = None):
    # remove zero length tsvs
    tsv_list = list(filter(lambda x: x.num_cols() > 0, tsv_list))

    # base condition
    if (len(tsv_list) == 0):
        raise Exception("List of tsv is empty")

    # boundary condition
    if (len(tsv_list) == 1):
        return tsv_list[0]

    # get the first header
    header_fields = tsv_list[0].get_header_fields()

    # some debugging
    diff_cols = get_diffs_in_headers(tsv_list)
    same_cols = []
    for h in header_fields:
        if (h not in diff_cols):
            same_cols.append(h)

    # print if number of unique headers are more than 1
    if (len(diff_cols) > 0):
        # debug
        utils.debug("merge_intersect: missing columns: {}".format(str(diff_cols)[0:100] + "..."))

        # check which of the columns among the diff have default values
        if (def_val_map is not None):
            # create effective map with empty string as default value
            effective_def_val_map = {}

            # some validation. the default value columns should exist somewhere
            for h in def_val_map.keys():
                # check if all default columns exist
                if (h not in diff_cols and h not in same_cols):
                    raise Exception("Default value for a column given which does not exist:", h)

            # assign empty string to the columns for which default value was not defined
            for h in diff_cols:
                if (h in def_val_map.keys()):
                    utils.trace_once("merge_intersect: assigning default value for {}: {}".format(h, def_val_map[h]))
                    effective_def_val_map[h] = str(def_val_map[h])
                else:
                    utils.trace_once("merge_intersect: assigning empty string as default value to column: {}".format(h))
                    effective_def_val_map[h] = ""

            # get the list of keys in order
            keys_order = []
            for h in header_fields:
                keys_order.append(h)

            # append the missing columns
            for h in diff_cols:
                if (h not in header_fields):
                    keys_order.append(h)

            # create a list of new tsvs
            new_tsvs = []
            for t in tsv_list:
                # TODO: use better design
                t1 = t
                if (def_val_map is not None and len(def_val_map) > 0):
                    for d in diff_cols:
                        t1 = t1.add_const_if_missing(d, effective_def_val_map[d])
                else:
                    t1 = t1.add_empty_cols_if_missing(diff_cols)
                # append to tsv list
                new_tsvs.append(t1.select(keys_order))

            # return after merging. dont call merge recursively as thats a bad design
            return new_tsvs[0].union(new_tsvs[1:]) 
        else:
            # handle boundary condition of no matching cols
            if (len(same_cols) == 0):
                return tsv.create_empty()
            else:
                # create a list of new tsvs
                new_tsvs = []
                for t in tsv_list:
                    new_tsvs.append(t.select(same_cols))

                return new_tsvs[0].union(new_tsvs[1:])
    else:
        # probably landed here because of mismatch in headers position
        tsv_list2 = []
        for t in tsv_list:
            tsv_list2.append(t.select(same_cols))
        return merge(tsv_list2)

def read(input_file_or_files, sep = None, def_val_map = None, username = None, password = None, num_par = 0, s3_region = None, aws_profile = None):
    # convert the input to array
    input_files = utils.get_argument_as_array(input_file_or_files)

    # tasks 
    tasks = []

    def __read_inner__(input_file):
        # read file content                               
        lines = file_paths_util.read_file_content_as_lines(input_file, s3_region, aws_profile)
                                                          
        # take header and dat
        header = lines[0]
        data = lines[1:]
            
        # check if a custom separator is defined
        if (sep is not None):
            # check for validation
            for line in lines:
                if ("\t" in line):
                    raise Exception("Cant parse non tab separated file as it contains tab character:", input_file)

            # create header and data
            header = header.replace(sep, "\t")
            data = [x.replace(sep, "\t") for x in data]

        # return
        return tsv.TSV(header, data)

    # create tasks
    for input_file in input_files:
        # check if it is a file or url
        if (input_file.startswith("http")):
            tsv_list.append(read_url_as_tsv(input_file, username = username, password = password))
            tasks.append(utils.ThreadPoolTask(read_url_as_tsv, input_file, username = username, password = password))
        else:
            tasks.append(utils.ThreadPoolTask(__read_inner__, input_file))

    # get result
    tsv_list = utils.run_with_thread_pool(tasks, num_par = num_par, wait_sec = 1)

    # merge and return
    return merge(tsv_list, def_val_map = def_val_map)

def __read_with_filter_transform_select_func__(cols):
    # create a inner function
    def __read_with_filter_transform_select_func_inner__(mp):
        result_mp = {}
        for c in cols:
            if (c in mp.keys()):
                result_mp[c] = str(mp[c])

        # return
        return result_mp

    return __read_with_filter_transform_select_func_inner__

def read_with_filter_transform(input_file_or_files, sep = None, def_val_map = None, filter_transform_func = None, cols = None, transform_func = None, s3_region = None, aws_profile = None):
    # check if cols is defined
    if (filter_transform_func is not None and cols is not None):
        raise Exception("tsvutils: read_with_filter_transform: either of filter_transform_func or cols parameter can be used")

    # use the map function for cols
    if (cols is not None and len(cols) > 0):
        filter_transform_func = __read_with_filter_transform_select_func__(cols)

    # check if filter_transform_func is defined
    if (filter_transform_func is None):
        xtsv = read(input_file_or_files, sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)

        # apply transform_func if defined
        xtsv_transform = transform_func(xtsv) if (transform_func is not None) else xtsv

        # return
        return xtsv_transform
    else:
        # resolve input
        input_files = utils.get_argument_as_array(input_file_or_files)

        # initialize result
        tsv_list = []

        # common keys
        common_keys = {}

        # iterate over all input files
        for input_file in input_files:
            # read the file
            x = read(input_file, sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)

            # update the common
            for h in x.get_header_fields():
                if (h not in common_keys.keys()):
                    common_keys[h] = 0
                common_keys[h] = common_keys[h] + 1

            # gather maps of maps
            result_maps = []
            keys = {}

            # iterate over the records of each map and generate a new one
            for mp in x.to_maps():
                mp2 = filter_transform_func(mp)
                if (mp2 is not None):
                    if (len(mp2) > 0):
                        result_maps.append(mp2)
                    for k in mp2.keys():
                        keys[k] = 1

            # check for empty maps
            if (len(keys) > 0):
                # output keys
                keys_sorted = []
                first_file = read(input_files[0], sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)
                for h in first_file.get_header_fields():
                    if (h in keys.keys()):
                        keys_sorted.append(h)

                # new header and data
                header2 = "\t".join(keys_sorted)
                data2 = []

                # iterate and generate header and data
                for mp in result_maps:
                    fields = []
                    for k in keys_sorted:
                        fields.append(mp[k])
                    data2.append("\t".join(fields))

                # debugging
                utils.trace("tsvutils: read_with_filter_transform: file read: {}, after filter num_rows: {}".format(input_file, len(data2)))

                # result tsv
                xtsv = tsv.TSV(header2, data2)

                # apply transformation function if defined
                xtsv_transform = transform_func(xtsv) if (transform_func is not None) else xtsv
                tsv_list.append(xtsv_transform)

        # Do a final check to see if all tsvs are empty
        if (len(tsv_list) > 0):
            # call merge on tsv_list
            return merge(tsv_list)
        else:
            # create an empty data tsv file with common header fields
            header_fields = []
            first_file = read(input_files[0], sep = sep, def_val_map = def_val_map, s3_region = s3_region, aws_profile = aws_profile)
            for h in first_file.get_header_fields():
                if (common_keys[h] == len(input_files)):
                    header_fields.append(h)

            new_header = "\t".join(header_fields)
            new_data = []

            # return
            return tsv.TSV(new_header, new_data)

# TODO: replace this by etl_ext
def read_by_date_range(path, start_date_str, end_date_str, prefix, s3_region = None, aws_profile = None, granularity = "daily"):
    utils.warn_once("read_by_date_range: probably Deprecated")
    # read filepaths
    filepaths = file_paths_util.read_filepaths(path, start_date_str, end_date_str, prefix, s3_region, aws_profile, granularity)

    # check for headers validity
    if (file_paths_util.has_same_headers(filepaths, s3_region, aws_profile) == False):
        utils.warn("Mismatch in headers for different days. Choose the right date range: start: {}, end: {}".format(start_date_str, end_date_str))
        return None

    # read individual tsvs
    tsv_list = []
    for filepath in filepaths:
        x = read(filepath)
        tsv_list.append(x)

    # combine all together
    if (len(tsv_list) == 0):
        return None
    elif (len(tsv_list) == 1):
        return tsv_list[0]
    else:
        return tsv_list[0].union(tsv_list[1:])

def load_from_dir(path, start_date_str, end_date_str, prefix, s3_region = None, aws_profile = None, granularity = "daily"):
    # read filepaths
    filepaths = file_paths_util.read_filepaths(path, start_date_str, end_date_str, prefix, s3_region, aws_profile, granularity)
    return load_from_files(filepaths, s3_region, aws_profile)

def load_from_files(filepaths, s3_region, aws_profile):
    # check for headers validity
    if (file_paths_util.has_same_headers(filepaths, s3_region, aws_profile) == False):
        print("Invalid headers.")
        return None

    # initialize the file reader
    file_reader = file_paths_data_reader.FilePathsDataReader(filepaths, s3_region, aws_profile)

    # get header
    header = file_reader.get_header()
    data = []

    # get data
    while file_reader.has_next():
        # read next record
        line = file_reader.next()
        data.append(line)

    # close
    file_reader.close()

    return tsv.TSV(header, data)

# TODO: use explode_json
def load_from_array_of_map(map_arr):
    # take a union of all keys
    keys = {}

    # copy map array and remove any white spaces
    map_arr2 = []
    for mp in map_arr:
        mp2 = {}
        for k in mp.keys():
            # for robustness remove any special white space characters
            k2 = utils.replace_spl_white_spaces_with_space(k)
            v = mp[k]

            # check for the type of value
            if (isinstance(v, (str))):
                v2 = utils.replace_spl_white_spaces_with_space(v)
            elif (isinstance(v, (list))):
                v2 = ",".join(list([utils.replace_spl_white_spaces_with_space(t1) for t1 in v]))
            elif (isinstance(v, (dict))):
                v2 = utils.url_encode(json.dumps(v))
                k2 = "{}:json_encoded".format(k2)
            elif (isinstance(v, (int))):
                v2 = str(v)
            else:
                v2 = str(v)

            # assign to map 
            mp2[k2] = v2

        # append
        map_arr2.append(mp2)

    # iterate over all maps
    for mp in map_arr2:
        for k in mp.keys():
            keys[k] = 1

    # sort the keys alphabetically
    sorted_keys = sorted(list(set(keys.keys())))

    # create header
    header = "\t".join(sorted_keys)
    header_fields = header.split("\t")
    header_map = {}
    for i in range(len(header_fields)):
        h = header_fields[i]
        header_map[h] = i

    data = []
    # create data
    for mp in map_arr2:
        fields = []
        for k in header_fields:
            v = ""
            if (k in mp.keys()):
                # read as string and replace any newline or tab characters
                v = utils.replace_spl_white_spaces_with_space(mp[k])

            # append
            fields.append(v)

        # create line
        line = "\t".join(fields)
        data.append(line)

    # create tsv
    return tsv.TSV(header, data).validate()

def save_to_file(*args, **kwargs):
    utils.noop_after_n_warnings("Deprecated. Use omigo_hydra instead", hydra.save_to_file, *args, **kwargs)

def check_exists(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use omigo_hydra instead")
    return hydra.check_exists(*args, **kwargs)

# Deprecated methods. Use wsclient package instead
def read_url_json(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use wsclient instead", num_warnings = 100, sleep_secs = 300)
    return wsclient.read_url_json(*args, **kwargs)

def read_url_response(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use wsclient instead", num_warnings = 100, sleep_secs = 300)
    return wsclient.read_url_response(*args, **kwargs)

def read_url(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use wsclient instead", num_warnings = 100, sleep_secs = 300)
    return wsclient.read_url(*args, **kwargs)

def read_url_as_tsv(*args, **kwargs):
    utils.rate_limit_after_n_warnings("Deprecated. Use wsclient instead", num_warnings = 100, sleep_secs = 300)
    return wsclient.read_url_as_tsv(*args, **kwargs)


